﻿using System.Collections.Generic;
using UnityEngine;
using UnityEngine.SceneManagement;
using System;
using System.Runtime.Serialization.Formatters.Binary;
using System.IO;

public class GameManager : MonoBehaviour {
	[Serializable]
	public class GameData{
		public Dictionary<string, float> levelTimes;
		public float[] totalTimes;
	}

	private static GameManager _instance = null;
	public GameData savedGameData;
	public Dictionary<string, float> currentLevelTimes;
	public float currentTotalTime;

	public Scene[] GameLevels;

	// Use this for initialization
	void Awake () {
		if(_instance == null){			
			DontDestroyOnLoad(gameObject);
			_instance = this;
		} else if(_instance != this) {
			Destroy(gameObject);
		}
	}

	void Start(){
		savedGameData = GameManager.LoadData();
		print("Loaded game data!");
		GameManager.AuditSavedData();

		SceneManager.activeSceneChanged += OnActiveSceneChanged;
	}

	private static void AuditSavedData(){
		print("BEST LEVEL TIMES:");
		foreach(KeyValuePair<string, float> entry in _instance.savedGameData.levelTimes){
			string levelName = entry.Key;
			float levelBestTime = entry.Value;
			Debug.LogFormat("  Level “{0}” -- {1}", levelName, levelBestTime);
		}

		print("TOP TOTAL TIMES:");
		for(int i=0; i<_instance.savedGameData.totalTimes.Length; i++){
			Debug.LogFormat("  {0}", _instance.savedGameData.totalTimes[i]);
		}
	}

	void OnActiveSceneChanged(Scene OldScene, Scene NewScene){
		string[] MouseScenes = { "Title Screen", "End Game" };

		if(Array.IndexOf(MouseScenes, NewScene.name) >= 0){
			Cursor.lockState = CursorLockMode.None;
			Cursor.visible = true;
		}
	}

	public static void StartNewGame(){
		_instance.currentLevelTimes = new Dictionary<string, float>();
		_instance.currentTotalTime = 0;
		GoToLevel("Level 1");
	}

	public static void GoToNextLevel(){
		Scene ActiveScene = SceneManager.GetActiveScene();
		int NextScene = (ActiveScene.buildIndex + 1) % SceneManager.sceneCountInBuildSettings;
		SceneManager.LoadScene(NextScene);
	}

	public static void GoToLevel(string levelName){
		SceneManager.LoadScene(levelName);
	}

	public static void CompleteLevel(string levelName, float levelTime){
		Debug.LogFormat("Finished {0} with a time of {1}!", levelName, levelTime);

		if(_instance.currentLevelTimes != null){
			_instance.currentLevelTimes[levelName] = levelTime;
			if(levelTime < GetBestTimeForLevel(levelName)){
				Debug.LogFormat(" THAT'S A NEW RECORD!");
				SetBestTimeForLevel(levelName, levelTime);
			}
		}

		GoToNextLevel();
	}

	public static float GetTotalTime(){
		float ret = 0;
		if(_instance.currentLevelTimes != null){
			foreach(KeyValuePair<string, float> entry in _instance.currentLevelTimes){
				// do something with entry.Value or entry.Key
				ret += entry.Value;
			}
		}

		return ret;
	}

	public static Dictionary<string, float> GetLevelTimes(){
		return _instance.currentLevelTimes;
	}

	public static float GetBestTimeForLevel(string levelName){
		float ret = 6099.99f;
		if(_instance.savedGameData.levelTimes.ContainsKey(levelName)){
			ret = _instance.savedGameData.levelTimes[levelName];
		}

		return ret;
	}

	public static float[] GetBestTotalTimes(){
		return _instance.savedGameData.totalTimes;
	}

	public static void SetBestTotalTimes(float[] Times){
		Debug.LogFormat("Setting High Score list to: {0}", Times);
		_instance.savedGameData.totalTimes = Times;
		SaveData();
	}

	public static float SetBestTimeForLevel(string levelName, float time){
		float minTime = 0f;
		float maxTime = 6099.99f;
		Debug.LogFormat("Set best time for {0} to {1}", levelName, time);

		if(_instance.savedGameData.levelTimes.ContainsKey(levelName)){
			maxTime = _instance.savedGameData.levelTimes[levelName];
		}

		float bestTime = Mathf.Clamp(time, minTime, maxTime);
		_instance.savedGameData.levelTimes[levelName] = bestTime;

		SaveData();

		return bestTime;
	}

	public static void EraseSaveData(){
		_instance.savedGameData.levelTimes = new Dictionary<string, float>();
		_instance.savedGameData.totalTimes = new float[3]{ 
			6099.99f, 
			6099.99f, 
			6099.99f 
		};

		SaveData();
	}

	public static void SaveData(){
		Debug.LogFormat("Saving Data...");
		AuditSavedData();
		BinaryFormatter bf = new BinaryFormatter();
		FileStream file = File.Create(Application.persistentDataPath + "/gameData.dat");
		GameData dat = new GameData();

		dat.levelTimes = _instance.savedGameData.levelTimes;
		dat.totalTimes = _instance.savedGameData.totalTimes;
		bf.Serialize(file, dat);
		file.Close();
	}

	public static GameData LoadData(){
		GameData ret = new GameData();

		if(File.Exists(Application.persistentDataPath + "/gameData.dat")){
			BinaryFormatter bf = new BinaryFormatter();
			FileStream file = File.Open(Application.persistentDataPath + "/gameData.dat", FileMode.Open);
			ret = (GameData)bf.Deserialize(file);
			file.Close();

			Debug.LogFormat("loaded data from {0}", Application.persistentDataPath + "/gameData.dat");
		} else {
			ret.levelTimes = new Dictionary<string, float>();
			ret.totalTimes = new float[3]{ 
				6099.99f, 
				6099.99f, 
				6099.99f
			};
		}

		return ret;
	}
}
